# -*- coding: utf-8 -*-

from __future__ import print_function
import os, re, sys, json, requests
from functools import wraps
from . import base
from .base import log, pjoin

vars_re = r"\{%-? *do *vars\.__setitem__\( *'sample_id' *, *(\d+) *\) -%\}"

title_choices = {
	'sample' : {
		u'样例 *#?(\\d+)?',
		u'输入输出样例 *#?(\\d+)?',
		u'样例输入输出 *#?(\\d+)?',
		u'测试用例 *#?(\\d+)?',
		r'sample\s* *#?(\d+)?',
		vars_re
	},
	'sample input' : {
		u'样例 *(\\d+)? *输入 *#?(\\d+)?',
		u'输入 *(\\d+)? *样例 *#?(\\d+)?',
		r'sample\s*input\s*(\d+)?',
		r'input\s*sample\s*(\d+)?'
	},
	'sample output' : {
		u'样例 *(\\d+)? *输出 *#?(\\d+)?',
		u'输出 *(\\d+)? *样例 *#?(\\d+)?',
		r'sample\s*output\s*(\d+)?',
		r'output\s*sample\s*(\d+)?'
	},
	'sample explanation' : {
		u'样例(\\d+)?说明(\\d+)?',
		u'样例(\\d+)?解释(\\d+)?',
		r'sample\s*explanation\s*(\d+)?'
	},
	'input' : {
		u'输入',
		'input'
	},
	'output' : {
		u'输出',
		'output'
	},
	'explanation' : {
		u'解释',
		u'说明',
		'explanation'
	},
	'input format' : {
		u'输入格式',
		u'输入说明',
		u'输入要求',
		u'输入描述',
		'input format'
	},
	'output format' : {
		u'输出格式',
		u'输出说明',
		u'输出要求',
		u'输出描述',
		'output format'
	},
	'description' : {
		u'题目描述',
		u'题目说明',
		u'题目',
		u'问题描述',
		u'问题说明',
		u'问题',
		'description',
		'statement',
	},
	'background' : {
		u'题目背景',
		u'背景',
		'background',
	},
	'hint' : {
		u'提示',
		u'温馨提示',
		u'hint'
	},
	'subtasks' : {
		u'(数据规模|数据范围|要求|限制|约束|约定|子任务)((和|与)(数据规模|数据范围|要求|限制|约束|约定|子任务|说明))?',
		r'subtask(s?)'
	},
	'scoring' : {
		u'评分方式',
		'scoring'
	}
}

new_re_symbols = {
	r'\C' : '[\u4e00-\u9fbf]',
	r'\P' : '[\uff00-\uffef\u3000-\u303f\u2000-\u203f]'
}

text_patterns = {
	u'中文和英文、字符串或公式中间没有空格' : r'\C(?:\w|`|\$)',
	u'英文、字符串或公式和中文中间没有空格' : r'(?:\w|`|\$)\C',
	u'中文之间有空格' : r'\C \C',
	u'使用了英文标点或不在公式中的运算符号' : '[,\\.\\?\\:;\'"\\[\\]\\{\\}!\\\\\\|\\(\\)\\+=]'
}

title_re = re.compile('^(' + '|'.join(['|'.join(value) for value in title_choices.values()]) + ')$')
std_title_re = re.compile('^\{\{ *s\( *\'(.*)\' *(( *, *[-+.\'\"a-zA-Z0-9_ ]+)*)\) *\}\} *$')
doc_title_re = u'^#* *( *【|\{\{ *_ *\( *\')?(%s)(\' *\) *\}\}|】)?( *\{#.*\})? *#* *\\d*$'
html_equation_re = re.compile(r'\\\[(.*)\\\]\{\.math \.inline\}')
format_log_re = re.compile(r'^(\w) (\d+) (.*)$')
itemize_re = re.compile(r'^\d+\.|\*|\-|\+ ')
md_img_re = re.compile(r'^\!\[.*\]\((.*)\)$')
html_img_re = re.compile(r'^<img.*src="(.*)".*/?>$')

math_trans = {
	'*' : '',
	u'≤' : r' \le',
	u'≥' : r' \ge',
	u'←' : r' \leftarrow',
	u'→' : r' \rightarrow',
	u'×' : r' \times',
	u'∑' : r' \sum',
	r'\$' : ''
}

def is_digit(s):
	for c in s:
		if not '0' <= c <= '9':
			return False
	return True

def is_english(s):
	for c in s:
		if not('a' <= c <= 'z' and 'A' <= c <= 'Z'):
			return False
	return True

def is_chinese(s):
	for c in s:
		if not u'\u4e00' <= c <= u'\u9fbf':
			return False
	return True

def get_text(s):
	ret = ''
	for c in s:
		if is_chinese(c) or is_english(c) or is_digit(c):
			ret += c
	return ret.lower()
	
def html_math2tex_math(s):
	ret = ''
	i = 0
	l = len(s)
	in_sup = False
	in_sub = False
	while True:
		if i == l:
			break
		if i + 1 < l:
			flag = False
			if s[i:i+2] == '\*':
				ret += '*'
			elif s[i:i+2] == '\^':
				if in_sup:
					ret += ' }'
				else:
					ret += '^{ '
				in_sup ^= True
			elif s[i:i+2] == '\~':
				if in_sub:
					ret += ' }'
				else:
					ret += '_{ '
				in_sub ^= True
			elif s[i:i+2] in math_trans:
				ret += math_trans[s[i:i+2]]
			else:
				flag = True
			if not flag:
				i += 2
				continue
		if s[i] in math_trans:
			ret += math_trans[s[i]]
		else:
			ret += s[i]
		i += 1
	return ret

class Section:
	used = {'input' : 0, 'output' : 0}
	def __init__(self, name = None, args = None):
		self.name = name
		if args == None:
			self.args = []
		else:
			self.args = args
		self.lines = []
	def clear_paragraph(self):
		flag = True
		for idx, line in enumerate(self.lines):
			if idx % 2 == 1 and line.rstrip('\n\r') != '':
				flag = False
		if flag:
			self.lines = [line for idx, line in enumerate(self.lines) if idx % 2 == 0]
	def write(self, f, idx = None):
		if self.name == 'sample':
			if len(self.lines) == 0:
				return
			else:
				self.args.append('')
		for name, suf in [('input', '.in'), ('output', '.ans')]:
			if self.name == 'sample ' + name or self.name == name:
				self.used[name] += 1
				self.clear_paragraph()
				with open(pjoin(base.prob.path, 'down', str(self.used[name]) + suf), 'w') as ff:
					if len(self.lines) > 1 and self.lines[0].startswith('```'):
						lines = self.lines[1:-1]
					else:
						lines = self.lines
					for line in lines:
						ff.write(line + '\n')
				if name == 'input':
					f.write(b'{{ s(\'sample\', %d) }}\n\n{{ self.sample_text() }}\n\n' % self.used[name])
				return
		if idx == 0:
			f.write(b'{{ self.title() }}\n')
		if idx == 0 and self.name and self.name not in title_choices:
			base.prob['title'][base.prob.lang()] = self.name.strip()
		elif self.name:
			f.write(("\n{{ s('%s'%s) }}\n\n" % (
				self.name,
				', ' + ', '.join(map(json.dumps, self.args)) if len(self.args) > 0 else ''
			)).encode('utf-8'))
		for line in self.lines:
			f.write((line + '\n').encode('utf-8'))
	def is_empty(self):
		return not self.name and len(self.lines) == 0
	def format_line(self, line):
		ret = line
		while True:
			m = html_equation_re.search(ret)
			if not m:
				break
			while True:
				sm = m
				l, r = sm.span()
				sub = sm.string[l:r - 1]
				m = html_equation_re.search(sub)
				if m:
					continue
				sub = sm.string[l:r]
				res = sm.group(1)
				ret = ret.replace(sub, '$%s$' % html_math2tex_math(res))
				break
		return ret
	def format(self):
		self.lines = [self.format_line(line) for line in self.lines]

def sure_title(line):
	inp = line
	if inp.startswith('##'):
		return True
	if inp.startswith(u'【') and inp.endswith(u'】'):
		return True
	if std_title_re.match(inp):
		return True
	if re.match(vars_re, inp):
		return True
	txt = get_text(inp)
	if title_re.match(txt):
		return True
	return False

def get_title(line):
	m = std_title_re.match(line)
	if m:
		gps = m.groups()
		name = gps[0]
		args = json.loads('[' + ','.join(gps[1].split(',')[1:]) + ']')
		return name, args
	m = re.match(doc_title_re % '[-+.\'\"a-zA-Z0-9_ \u4e00-\u9fa5\u3040-\u309f\u30a0-\u30ff]+', line.lower())
	if m:
		title = m.group(2)
		m = title_re.match(title)
		if m:
			for key, pts in title_choices.items():
				for pt in pts:
					m = re.match('^' + pt + '$', title)
					if m:
						name = key
						args = []
						if key.startswith('sample'):
							for s in m.groups():
								if s and s != '':
									args = [json.loads(s)]
									break
						log.info(f"标题：{name} {args}")
						return name, args
		return title, []
	return line, []

def to_sections(lines):
	def clear(buff):
		if buff != []:
			cur.lines.append(' '.join(buff))
			cur.lines.append('')
			buff = []
		return buff
	sections = []
	cur = Section()
	buff = []
	in_quote = False
	in_quote_space = False
	in_quote_tab = False
	in_equation = False
	in_table = False
	last_title = 0
	tbl_lino = 0
	tbl_no = 0
	img_no = 0
	for line in lines:
		if in_quote_space:
			if line.startswith('    '):
				cur.lines.append(line.lstrip(' '))
				continue
			else:
				in_quote_space = False
				cur.lines.append('```')
		elif in_quote_tab:
			if line.startswith('\t'):
				cur.lines.append(line.lstrip(' '))
				continue
			else:
				in_quote_tab = False
				cur.lines.append('```')
		elif in_table:
			if line.startswith('  '):
				cur.lines.append(line)
				continue
			else:
				in_table = False
		if in_quote:
			if line == '```':
				in_quote = False
			cur.lines.append(line)
		elif in_equation:
			if line == '$$':
				in_equation = False
			cur.lines.append(line)
		else:
			last_title = last_title - 1 if last_title >= 1 else 0
			if line.startswith('|') and line.endswith('|'):
				if tbl_lino == 0:
					tbl_no += 1
					log.info(f"表格：{tbl_no}")
					buff = clear(buff)
					cur.lines.append("{{ tbl('%d') }}\n" % tbl_no)
					base.mkdir('tables')
					ftbl = open(f"tables/{tbl_no}.pyinc", 'w', encoding = 'utf-8')
					ftbl.write('return [\n')
				tbl_lino += 1
				if tbl_lino != 2:
					ftbl.write("\t[r'" + "', r'".join(map(lambda x : x.strip(), line.strip('|').split('|'))) + "'],\n")
				continue
			elif tbl_lino != 0:
				ftbl.write(']\n')
				ftbl.close()
				tbl_lino = 0
			img_match = md_img_re.match(line.strip()) or html_img_re.match(line.strip())
			if line.startswith('```'):
				in_quote = True
				buff = clear(buff)
				cur.lines.append(line)
			elif line.strip() == '$$':
				in_equation = True
				buff = clear(buff)
				cur.lines.append(line)
			elif sure_title(line):
				buff = clear(buff)
				title, args = get_title(line)
				if not cur.is_empty():
					sections.append(cur)
				cur = Section(title, args)
				last_title = 2
			elif line.startswith('    '):
				in_quote_space = True
				buff = clear(buff)
				cur.lines.append('```')
				cur.lines.append(line.lstrip(' '))
			elif line.startswith('\t'):
				in_quote_space = True
				buff = clear(buff)
				cur.lines.append('```')
				cur.lines.append(line.lstrip(' '))
			elif line.startswith('  '):
				in_table = True
				buff = clear(buff)
				cur.lines.append(line)
			elif line.startswith('------'):
				if last_title >= 1:
					continue
				if buff != []:
					buff.pop()
				buff = clear(buff)
				title, args = get_title(last)
				if not cur.is_empty():
					sections.append(cur)
				cur = Section(title, args)
			elif itemize_re.match(line.lstrip()):
				buff = clear(buff)
				cur.lines.append(line)
			elif line == '':
				buff = clear(buff)
			elif img_match:
				img_url = img_match.group(1)
				img_suf = img_url.split('.')[-1]
				img_no += 1
				buff = clear(buff)
				cur.lines.append("{{ img('%s.%s') }}\n" % (img_no, img_suf))
				log.info(f"图片：{img_no}.{img_suf} {img_url}")
				try:
					r = requests.get(img_url)
					base.mkdir('resources')
					open(f"resources/{img_no}.{img_suf}", 'wb').write(r.content)
				except Exception as e:
					log.warning(f'从网站上下载图片失败，请手动下载并添加到resources/{img_no}.{img_suf}')
					log.info(e)
			else:
				buff.append(line)
		last = line
	buff = clear(buff)
	if tbl_lino != 0:
		ftbl.write(']\n')
		ftbl.close()
		tbl_lino = 0
	if not cur.is_empty():
		sections.append(cur)
	return sections

def format():
	for base.prob in base.probs():
		prob = base.prob
		path = pjoin(base.conf.path, 'statement', base.conf.lang() + '.md')
		lines = [line.rstrip(b'\r\n').decode('utf-8') for line in open(path, 'rb').readlines()]
		sections = to_sections(lines)
		for section in sections:
			section.format()
		with open(path, 'wb') as f:
			for idx, section in enumerate(sections):
				section.write(f, idx)
	base.save_json(base.conf)

def format_check_one(path):
	log.info(u'检查文件`%s`。' % path)
	widths = [
		(126, 1), (159, 0), (687, 1), (710, 0), (711, 1), 
		(727, 0), (733, 1), (879, 0), (1154, 1), (1161, 0), 
		(4347, 1), (4447, 2), (7467, 1), (7521, 0), (8369, 1), 
		(8426, 0), (9000, 1), (9002, 2), (11021, 1), (12350, 2), 
		(12351, 1), (12438, 2), (12442, 0), (19893, 2), (19967, 1),
		(55203, 2), (63743, 1), (64106, 2), (65039, 1), (65059, 0),
		(65131, 2), (65279, 1), (65376, 2), (65500, 1), (65510, 2),
		(120831, 1), (262141, 2), (1114109, 1),
	]
	def get_width(o):
		if o == '\t':
			return 4
		o = ord(o)
		if o == 0xe or o == 0xf:
			return 0
		for num, wid in widths:
			if o <= num:
				return wid
		return 1
	def output():
		level = {
			'E' : log.error,
			'W' : log.warning,
			'I' : log.info,
			'D' : log.debug,
		}[logs[idx][0]]
		level(u'第%d行，第%d个字：%s' % (lineno, col, logs[idx][2]))
		lef = col - 10 if col > 10 else 0
		rig = min(lef + 21, len(code.rstrip()))
		log.info(code[lef:rig].replace('\u200b', '?').replace('\t', '    '))
		lef_cur = 0
		for c in code[:lef]:
			lef_cur += get_width(c)
		log.info(' ' * (cur - lef_cur) + u'↑')
	os.system('%s < %s > format.tmp 2> format.log' % (
		pjoin(base.tool_path, base.format_checker_name),
		path
	))
	tot = 0
	logs = []
	for line in open('format.log', 'rb'):
		m = format_log_re.match(line.decode('utf-8').rstrip())
		if m:
			logs.append((m.group(1), int(m.group(2)), m.group(3)))
	if len(logs) == 0:
		return
	idx = 0
	for lineno, line in enumerate(open(path, 'rb'), 1):
		code = line.decode('utf-8')
		cur = 0
		for col, ch in enumerate(code):
			tot += len(ch.encode('utf-8'))
			if tot >= logs[idx][1]:
				output()
				idx += 1
			if idx >= len(logs):
				return
			cur += get_width(ch)

def format_check():
	base.check_install('format')
	for base.prob in base.probs():
		prob = base.prob
		path = pjoin(prob.path, 'statement', 'zh-cn.md')
		format_check_one(path)

def load():
	if len(base.args) != 1:
		log.error(u'无法转换，必须传入恰好一个参数。')
		sys.exit(1)
	if base.conf.folder != 'problem':
		log.error(u'只能处理单个题目，请到题目目录下进行操作。')
		sys.exit(1)
	os.system('pandoc %s -o %s' % (base.args[0], pjoin(base.conf.path, 'statement', base.conf.lang() + '.md')))

class_list = {
	'load' : load,
	'format' : format,
	'check' : format_check
}

if __name__ == '__main__':
	try:
		if base.init():
			base.check_install('pandoc')
			base.check_install('jinja2')
			for base.work in base.works:
				for base.lang in base.langs:
					base.run_exc(class_list[base.work])
		else:
			log.info(u'这是用来处理题面的工具。')
			log.info(u'支持的工作：')
			log.info(u'  load     必须包含一个参数输入文件名，表示要导入的题面。')
			log.info(u'  format   尝试对当前的中文题面进行简单的格式化，危险操作请注意备份。')
			log.info(u'  check    对题面进行格式检查。')
			sys.exit(1)
	except base.NoFileException as e:
		log.error(e)
		log.info(u'尝试使用`python -m tuack.gen -h`获取如何生成一个工程。')
		sys.exit(1)




# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import re
import sys
import json
import datetime
import shutil
import subprocess
import time
import signal
import zipfile
from multiprocessing import Process, Queue
from functools import wraps
from threading import Timer
from . import num2chinese
import platform
from . import base
from .base import log, pjoin

def find_all_data(kind, folder, key, conf = None):
	if not conf:
		conf = base.conf
	def find_data(path = ''):
		full_path = pjoin(prob.path, folder, path)
		for f in os.listdir(full_path):
			if os.path.isfile(pjoin(full_path, f)):
				if not f.endswith('.in'):
					continue
				fans = pjoin(full_path, f[:-3]) + '.ans'
				if not os.path.exists(fans) or not os.path.isfile(fans):
					continue
				name = base.rjoin(path, f[:-3])
				if name not in new_data and name not in exist_data:
					new_data.add(name)
			else:
				find_data(base.rjoin(path, f))
	def parse(data):
		tmp = base.sorter()(map(str, list(data)))
		ret = []
		for i in tmp:
			try:
				ret.append(int(i))
			except Exception as e:
				ret.append(i)
		return ret
	for prob in conf.probs(no_repeat = True):
		log.info(u'在题目`%s`下搜索%s数据。' % (prob.route, key))
		is_lfs = os.path.exists(pjoin(prob.path, '.gitattributes'))
		new_data = set()
		exist_data = set(map(str, prob.__getattribute__(key)))
		try:
			find_data()
		except Exception as e:
			log.warning(e)
		new_data = parse(new_data)
		for datum in new_data:
			log.info(u'发现新数据`%s`。' % (pjoin(prob.path, folder, str(datum))))
		if not is_lfs:
			for tdata in (exist_data, new_data):
				for datum in tdata:
					for suf in ('.in', '.out', '.ans'):
						fname = pjoin(prob.path, folder, str(datum) + suf)
						if os.path.exists(fname) and os.path.getsize(fname) >= 1024 * 1024:
							log.warning(u'题目`%s`下%s数据达到了1MiB，建议在该题下使用`python tuack.gen lfs`。' % (prob.route, key))
							break
					else:
						continue
					break
				else:
					continue
				break
		prob.__getattribute__(key).__iadd__(new_data)
		if kind not in prob:
			prob[kind] = []
		if len(new_data) > 0:
			prob[kind].append({'cases' : new_data})
		base.save_json(prob)
	
def find_all_code():
	def find_code(user, path):
		full_path = pjoin(prob.path, user, path)
		for f in os.listdir(full_path):
			if base.user_skip.match(f):
				continue
			if os.path.isfile(pjoin(full_path, f)):
				for key in base.compilers:
					if not f.endswith('.' + key):
						continue
					code_path = base.rjoin(user, path, f).replace('//', '/')
					if code_path not in exist_code:
						prob['users'][user][base.rjoin(path, f).replace('//', '/')] = {
							'path' : code_path,
							'expected' : '== 100' if 'std' in code_path or prob['name'] in code_path else []
						}
						log.info(u'发现新源代码`%s`。' % code_path)
					break
			else:
				find_code(user, base.rjoin(path, f))

	for prob in base.probs(no_repeat = True):
		if prob['type'] == 'output':
			log.info(u'题目`%s`是提交答案题，跳过。' % prob.route)
			continue
		log.info(u'在题目`%s`下搜索源代码。' % prob.route)
		if 'users' not in prob:
			prob['users'] = {}
		exist_code = set()
		for user, algos in prob['users'].items():
			for algo, path in algos.items():
				exist_code.add((path['path'] if type(path) == dict else path).replace('//', '/'))
		for f in os.listdir(prob.path):
			if base.user_skip.match(f) or os.path.isfile(pjoin(prob.path, f)):
				continue
			if f not in prob['users']:
				prob['users'][f] = {}
			find_code(f, '')
			if len(prob['users'][f]) == 0:
				prob['users'].pop(f)
		base.save_json(prob)

def sample_copy(src, tgt = None, path = '', base_path = 'sample'):
	if not tgt:
		tgt = src
	full_tgt = pjoin(path, tgt)
	if os.path.exists(full_tgt) and not os.path.isfile(full_tgt):
		full_tgt = pjoin(full_tgt, src)
	if not os.path.exists(full_tgt):
		log.info(u'生成文件`%s`' % full_tgt)
		base.copy(
			pjoin(base.path, base_path),
			src,
			full_tgt
		)

def new_dir(folder, args = None):
	if not args:
		args = base.args
	if len(args) == 0:
		dirs = ['.']
	else:
		if not base.conf:
			log.error(u'当前文件夹下没有找到合法的`conf.json`文件。')
			log.info(u'尝试使用`python -m tuack.gen -h`获取如何生成一个工程。')
			return
		if folder == 'contest':
			log.error(u'文件夹层次错误：比赛工程不能放在其他工程中。')
			return
		if folder == 'day' and base.conf.folder != 'contest':
			log.error(u'文件夹层次错误：比赛日工程只能放在比赛工程中。')
			return
		if (folder == 'problem' or folder == 'empty') and base.conf.folder != 'day':
			log.error(u'文件夹层次错误：题目工程只能放在比赛日工程中。')
			return
		dirs = args
	for path in dirs:
		if not os.path.exists(path):
			os.makedirs(path)
		if path != '.' and path in base.conf['subdir']:
			log.warning(u'`%s`已经在工程中了，跳过它的建立。' % path)
			log.info(u'如果需要强制建立，需要将其从`conf.yaml`中移除。')
			continue
		copy = lambda src, tgt = None: sample_copy(src, tgt, path)
		def copy_conf(inp):
			if base.dump_format != 'json':
				open(pjoin(path, 'conf.' + base.dump_format), 'wb').write(
					base.dump_formats[base.dump_format](
						json.loads(open(pjoin(base.path, 'sample', inp), 'rb').read().decode('utf-8'))
					)
				)
			else:
				copy(inp, 'conf.json')
		copy(folder + '.gitignore', '.gitignore')
		copy_conf(folder + '.json')
		if folder == 'problem' or folder == 'empty':
			if base.git_lfs:
				copy(folder + '.gitattributes', '.gitattributes')
			else:
				log.info(u'现在不默认用git-lfs，如需，用`python -m tuack.gen lfs`添加。')
			for ff in ('data', 'down', 'pre', 'statement', 'tables', 'resources', 'solution'):
				st_path = pjoin(path, ff)
				if not os.path.exists(st_path):
					os.makedirs(st_path)
				if not os.path.exists(pjoin(base.path, 'sample-' + folder, ff)):
					continue
				for f in os.listdir(pjoin(base.path, 'sample-' + folder, ff)):
					sample_copy(pjoin(ff, f), pjoin(ff, f), path, 'sample-' + folder)
			sample_copy('gen.py', 'gen.py', path, 'sample-problem')
		if path != '.':
			conf = base.load_json(path)
			conf['name'] = path
			conf.path = path
			if path == 'day0':
				conf['title']['zh-cn'] = u'试机'
			elif path[:3] == 'day':
				try:
					num = int(path[3:])
					conf['title']['zh-cn'] = u'第' + num2chinese.num2chinese(num) + u'试'
				except Exception as e:
					pass
			base.save_json(conf)
	if len(args) != 0:
		for arg in args:
			if arg not in base.conf['subdir']:
				base.conf['subdir'].append(arg)
		base.save_json(base.conf)

def upgrade():
	if base.conf:	#是conf.json格式的老版本
		used = {}
		def upgrade_r(conf):
			'''
			非最新版本return False
			'''
			def upgrade_None(conf):
				log.info(u'将`%s`从None版本升级到0版本。' % conf.route)
				conf['version'] = 0
				if conf.folder == 'problem':
					for folder, cases, key in [('data', 'test cases', 'data'), ('down', 'sample count', 'samples')]:
						if cases in conf:
							cnt = conf.pop(cases)
							for i in range(1, cnt + 1):
								for suf in ['.in', '.ans']:
									try:
										ff = pjoin(conf.path, folder, conf['name'] + str(i) + suf)
										ft = pjoin(conf.path, folder, str(i) + suf)
										os.rename(ff, ft)
									except FileNotFoundError as e:
										if os.path.exists(ft):
											log.info('`%s`已经改名，跳过重命名。' % ft)
										else:
											log.error('`%s`和`%s`都找不到，可能你的文件命名错误。' % (ff, ft))
							work_list[key](conf)
				return False
			def upgrade_0(conf):
				log.info(u'将`%s`从0版本升级到1版本。' % conf.route)
				conf['version'] = 1
				if conf.folder == 'problem':
					if 'users' not in conf:
						conf['users'] = {}
					if 'pre' not in conf:
						conf['pre'] = []
					for user, algos in conf['users'].items():
						for algo in algos:
							if type(algos[algo]) == str:
								algos[algo] = {'path' : algos[algo], 'expected' : {}}
				return False
			def upgrade_1(conf):
				log.info(u'将`%s`从1版本升级到2版本。' % conf.route)
				conf['version'] = 2
				if conf.folder == 'problem':
					if 'users' not in conf:
						conf['users'] = {}
					if 'packed' in conf:
						conf.pop('packed')
					for user, algos in conf['users'].items():
						for algo, info in algos.items():
							if type(info.get('expected', None)) == dict:
								info['expected'] = ['%s %s' % (key, val) for key, val in info['expected'].items()]
								if len(info['expected']) == 1:
									info['expected'] = info['expected'][0]
				return False
			def upgrade_2(conf):
				log.info(u'`%s`是最新版本。' % conf.route)
				return True
			path = os.path.abspath(conf.path)
			if path in used:
				for key in used[path]:
					conf[key] = used[path][key]
			while not eval('upgrade_' + str(conf['version']))(conf):
				pass
			for sub in conf.sub:
				upgrade_r(sub)
			used[path] = conf
		upgrade_r(base.conf)
		base.save_json(base.conf)
	else:			#是prob(s).json格式的老版本
		if not os.path.exists('probs.json'):
			log.error(u'找不到`probs.json`。')
			return
		old_json = json.loads(open('probs.json', 'rb').read().decode('utf-8'))
		new_dir('contest', [])
		base.conf = base.load_json()
		base.conf.pop('version')
		new_dir('day', base.sorter()(old_json.keys()))
		base.conf = base.load_json()
		for day in base.conf.sub:
			day.pop('version')
			day['subdir'] += [prob for prob in old_json[day['name']]]
		base.save_json(base.conf)
		base.conf = base.load_json()
		upgrade()

def copy_lfs():
	for prob in base.probs(no_repeat = True):
		copy = lambda src, tgt = None: sample_copy(src, tgt, prob.path)
		copy('problem.gitattributes', '.gitattributes')

def copy_chk():
	if base.conf['folder'] != 'problem':
		log.error(u'答案校验器只能添加给题目工程，请到题目目录下运行。')
		return
	for prob in base.probs(no_repeat = True):
		copy = lambda src, tgt = None: sample_copy(src, tgt, prob.path)
		if not os.path.exists(pjoin(prob.path, 'data', 'chk')):
			os.makedirs(pjoin(prob.path, 'data', 'chk'))
		copy('chk.cpp', pjoin('data', 'chk'))

def copy_prec():
	if base.conf['folder'] != 'contest':
		log.error(u'考生须知只能添加给比赛工程。')
		return
	path = '.'
	copy = lambda src, tgt = None: sample_copy(src, tgt, path)
	for ff in ('precautions', ):
		st_path = pjoin(path, ff)
		if not os.path.exists(st_path):
			os.makedirs(st_path)
		for f in os.listdir(pjoin(base.path, 'sample', ff)):
			copy(pjoin(ff, f), pjoin(ff, f))

def set_conf(params, folders, tp = str):
	if type(params) != tuple and type(params) != list:
		params = (params,)
	if base.conf['folder'] not in folders:
		log.error(f"只能在{folders}类型目录下运行当前命令")
		return
	for sub, val in zip(base.conf.sub, sys.argv[2+int(tp == json.loads):]):
		t = sub
		for i in params[:-1]:
			t = t[i]
		t[params[-1]] = tp(val)
		base.save_json(sub)

def set_length():
	if base.conf['folder'] not in {'day', 'contest'}:
		log.error(f"只能在{'day', 'contest'}类型目录下运行当前命令")
		return
	for sub, val in zip(base.conf.sub if base.conf['folder'] == 'contest' else [base.conf], sys.argv[2:]):
		sub['end time'] = (datetime.datetime.strptime(sub['start time'].split('+')[0], '%Y-%m-%d %H:%M:%S') + datetime.timedelta(hours = float(val))).strftime('%Y-%m-%d %H:%M:%S') + '+0800'
		base.save_json(sub)

work_list = {
	'data' : lambda conf = None: find_all_data('data', 'data', 'test_cases', conf),
	'samples' : lambda conf = None: find_all_data('samples', 'down', 'sample_cases', conf),
	'pre' : lambda conf = None: find_all_data('pre', 'pre', 'pre_cases', conf),
	'code' : find_all_code,
	'contest' : lambda : new_dir('contest'),
	'day' : lambda : new_dir('day'),
	'problem' : lambda : new_dir('problem'),
	'empty' : lambda : new_dir('empty'),
	'upgrade' : upgrade,
	'lfs' : copy_lfs,
	'chk' : copy_chk,
	'prec' : copy_prec,
	'title' : lambda : set_conf(('title', 'zh-cn'), {'contest', 'day'}),
	'time' : lambda : set_conf(('time limit', ), {'day'}, float),
	'memory' : lambda : set_conf(('memory limit', ), {'day'}, float),
	'conf' : lambda : set_conf(json.loads(sys.argv[2]), {'contest', 'day', 'problem'}, json.loads),
	'length' : set_length
}
work_list['auto'] = lambda : [work_list[key]() for key in ['data', 'samples', 'pre', 'code']]
work_list['n'] = work_list['contest']
work_list['d'] = work_list['day']
work_list['p'] = work_list['problem']
work_list['e'] = work_list['empty']
work_list['t'] = work_list['data']
work_list['s'] = work_list['samples']
work_list['u'] = work_list['upgrade']
work_list['a'] = work_list['auto']
work_list['l'] = work_list['title']
work_list['i'] = work_list['time']
work_list['m'] = work_list['memory']
work_list['g'] = work_list['length']
work_list['f'] = work_list['conf']

if __name__ == '__main__':
	try:
		result = base.init()
	except base.NoFileException as e:
		base.conf = None
		result = True
	if result:
		for base.work in base.works:
			base.run_exc(work_list[base.work])
	else:
		log.info(u'这个工具用于快速建立一道题目。')
		log.info(u'支持的工作：')
		log.info(u'  contest  在当前目录下生成一场比赛，不支持参数。')
		log.info(u'  n        同contest。')
		log.info(u'  day      无参数表示在当前目录下生成一个比赛日，比赛日可以是独立的工程；')
		log.info(u'           有参数表示在当前比赛下依次生成名叫参数1，参数2，…的比赛日，')
		log.info(u'           有参数必须保证当前目录是比赛。')
		log.info(u'  d        同day。')
		log.info(u'  problem  无参数表示在当前目录下生成一道题目，题目可以是独立的工程；')
		log.info(u'           有参数表示在当前比赛日下依次生成名叫参数1，参数2，…的题目，')
		log.info(u'           有参数必须保证当前目录是比赛日。')
		log.info(u'  p        同problem。')
		log.info(u'  empty    无参数表示在当前目录下生成一道无例子的题目，题目可以是独立的工程；')
		log.info(u'           有参数表示在当前比赛日下依次生成名叫参数1，参数2，…的题目，')
		log.info(u'           有参数必须保证当前目录是比赛日。')
		log.info(u'  e        同empty。')
		log.info(u'  data     在题目工程的data文件夹中搜索数据并添加到配置文件。')
		log.info(u'           对于比赛工程和比赛日工程，此操作将应用于所有子题目工程，下同。')
		log.info(u'  t        同data。')
		log.info(u'  samples  在题目工程的down文件夹中搜索样例并添加到配置文件。')
		log.info(u'  s        同samples。')
		log.info(u'  pre      在题目工程的pre文件夹中搜索预测试数据并添加到配置文件。')
		log.info(u'  code     在题目工程的非数据文件夹中搜索源代码并添加到配置文件。')
		log.info(u'  c        同code。')
		log.info(u'  auto     等于一次性做了上述4件事情。')
		log.info(u'  a        同auto。')
		log.info(u'  lfs      用git-lfs维护所有的*.in/out/ans，当数据较大时使用。')
		log.info(u'  chk      添加一个空的答案校验器或称spj，建议在此基础上修改以兼容。')
		log.info(u'  prec     对于比赛工程，添加考生须知文件。')
		log.info(u'  upgrade  升级老版本工程到现在版本的工程。')
		log.info(u'  u        同upgrade。')
		log.info(u'  title    设置每道题或每个比赛日的标题，依次传入标题名字。')
		log.info(u'  l        同title。')
		log.info(u'  time     设置每道题的时间限制，依次传入每道题的时间限制。')
		log.info(u'  i        同time。')
		log.info(u'  memory   设置每道题的空间限制，依次传入每道题的时间限制。')
		log.info(u'  m        同memory。')
		log.info(u'  length   设置每个比赛日或当前比赛日的时长，依次用浮点数以小时为单位传入。')
		log.info(u'  g        同length。')
		log.info(u'  conf     批量设置字段，先传入字段名（可用数组表示层次），然后依次用json格式传入要改的内容。例如：')
		log.info(u'           tuack.gen conf \'"file io"\' true true true true')
		log.info(u'           tuack.gen conf \'"compile"\' \'{"cpp" : "-O2 -std=c++14"}\' \'{"cpp" : "-O2 -std=c++14"}\'')
		log.info(u'           tuack.gen conf \'["compile","cpp"]\' \'"-O2 -std=c++14"\' \'"-O2 -std=c++14"\'')
		log.info(u'  f        同conf。')



# -*- coding: utf-8 -*-

from datetime import *
from math import *
import sys
if sys.version_info >= (3, 0): 
	from builtins import *
else:
	from __builtin__ import *
import json
from .num2chinese import *
from .base import log

def int_lg(num):
	if num == 0:
		return float('-inf')
	if num < 0:
		return float('nan')
	if num >= 10:
		if issubclass(type(num), int):
			return int_lg(num // 10) + 1
		else:
			return int_lg(num / 10) + 1
	if num < 1:
		return int_lg(num * 10) - 1
	return 0

def comma(num):
	if not issubclass(type(num), int):
		return str(num)
	if num < 0:
		return '-' + comma(-num)
	if num < 1000:
		return str(num)
	return comma(num // 1000) + ',' + '%03d' % (num % 1000)
	
def hn(num, style = None):
	'''
	适合阅读的数
	'''
	ret = ''
	l = 0
	if num == 0:
		return '0'
	if num < 0:
		neg = '-'
		num = -num
	else:
		neg = ''
	n = int_lg(num)
	if num != 10 ** n:
		if num // 10 ** n * 10 ** n == num:
			ret += str(num // 10 ** n)
			l += 1
		else:
			tmp = str(float(num) / 10 ** n)
			ret += tmp
			l += len(tmp)
		ret += ' \\times '
		l += 1
	ret += '10^{%d}' % n
	l += 2 + len(str(n))
	if style == 'x':
		ret = ret
	elif style == ',':
		ret = comma(num)
	else:
		ret = (comma(num) if l >= len(str(num)) * 4 // 3 else ret)
	return neg + ret
	
def js_hn(num, style = None):
	'''
	适合阅读的数，json版本
	'''
	return json.dumps(hn(num, style))[1:-1]

def cases(cases):
	ret = []
	st = None
	ed = None
	def deal():
		if st is None:
			return
		if issubclass(type(st), int) and st + 1 < ed:
			ret.append("%d \\sim %d" % (st, ed))
		else:
			ret.append(str(st))
			if st != ed:
				ret.append(str(ed))
	for i in cases:
		if issubclass(type(st), int) and issubclass(type(i), int) and ed + 1 == i:
			ed = i
		else:
			deal()
			st = i
			ed = i
	deal()
	return '$' + ','.join(ret) + '$'

def to_time(dttm):
	try:
		return datetime.strptime(dttm, '%Y-%m-%d %H:%M:%S')
		#return datetime.strptime(dttm, '%Y-%m-%d %H:%M:%S%z')	#python2 兼容性不好，暂时不用时区了
	except Exception as e:
		return datetime.strptime(dttm[:-5], '%Y-%m-%d %H:%M:%S')

def date(start, year = '-', month = '-', day = ''):
	st = datetime.strptime(start[:-5], '%Y-%m-%d %H:%M:%S')
	ret = st.strftime('%Y%%sQAQ%m%%sQAQ%d%%s').replace('QAQ0', '').replace('QAQ', '') % (year, month, day)
	return ret

def time_range(start, end, year = '-', month = '-', day = ''):
	st = datetime.strptime(start[:-5], '%Y-%m-%d %H:%M:%S')
	ed = datetime.strptime(end[:-5], '%Y-%m-%d %H:%M:%S')
	#st = datetime.strptime(start, '%Y-%m-%d %H:%M:%S%z')	#python2 兼容性不好，暂时不用时区了
	#ed = datetime.strptime(end, '%Y-%m-%d %H:%M:%S%z')
	ret = st.strftime('%Y%%sQAQ%m%%sQAQ%d%%s %H:%M').replace('QAQ0', '').replace('QAQ', '') % (year, month, day)
	if st.second:
		ret += ':%02d' % st.second
	ret += ' ~ '
	flag = False
	if flag or ed.year != st.year:
		flag = True
		ret += '%04d%s' % (ed.year, year)
	if flag or ed.month != st.month:
		flag = True
		ret += '%d%s' % (ed.month, month)
	if flag or ed.day != st.day:
		flag = True
		ret += '%d%s' % (ed.day, day)
	ret += ' %02d:%02d' % (ed.hour, ed.minute)
	if ed.second:
		ret += ':%02d' % ed.second
	return ret

def json_dumps_twice(s):
	return json.dumps(json.dumps(s))

class Int(int):
	def __new__(self, a):
		return super(Int, self).__new__(self, a)
	hn = lambda self, style = None: globals()['hn'](self, style)
	js_hn = lambda self, style = None: globals()['js_hn'](self, style)
	comma = lambda self: globals()['comma'](self)
	int_lg = lambda self: globals()['int_lg'](self)
	to_chinese = lambda self, big=False, simp=True, o=False, twoalt=False: num2chinese(self, big, simp, o, twoalt)

class Float(float):
	def __new__(self, a):
		return super(Float, self).__new__(self, a)
	hn = lambda self, style = None: globals()['hn'](self, style)
	js_hn = lambda self, style = None: globals()['js_hn'](self, style)
	comma = lambda self: globals()['comma'](self)
	int_lg = lambda self: globals()['int_lg'](self)
	to_chinese = lambda self, big=False, simp=True, o=False, twoalt=False: num2chinese(self, big, simp, o, twoalt)

class Args(object):
	def __init__(self, *data):
		self.data = data
	def get(self, key, d = None, skip_none = False):
		ret = []
		for i in self.data:
			if i is None:
				continue
			for j in (i if type(i) == list else [i]):
				v = j.get('args', {}).get(key)
				if not skip_none or v is not None:
					ret.append(v)
		return ret
	def keys(self):
		ky = set()
		for i in self.data:
			ky |= set(i.keys())
		return ky
	def __getitem__(self, key):
		return self.get(key)
	def sum(self, key, d = None):
		return sum(self.get(key, d, True))
	def min(self, key, d = None):
		return min(self.get(key, d, True))
	def max(self, key, d = None):
		return max(self.get(key, d, True))
	def prod(self, key, d = None):
		return prod(self.get(key, d, True))

a = Args
